﻿using System;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Diagnostics;
using System.IO;
using System.Windows;
using System.Windows.Input;
using System.Windows.Threading;
using System.Xml;
using Microsoft.Win32;
using mockupCirene.Helpers;
using mockupCirene.Model;

namespace mockupCirene.ViewModel
{

    /// <summary>
    /// This class implements the view model of the code instrumentation process.
    /// </summary>
    public class CodeInstrumentationViewModel : ObservableObject
    {
        private ObservableCollection<string> _errorData;

        /// <summary>
        /// Gets and sets error data during instrumentation process.
        /// </summary>
        public ObservableCollection<string> ErrorData
        {
            get
            {
                if (_errorData == null)
                {
                    ErrorData = new ObservableCollection<string>();
                }
                return _errorData;
            }
            set
            {
                _errorData = value;
                RaisePropertyChanged("ErrorData");
            }
        }

        private ObservableCollection<string> _outputData;

        /// <summary>
        /// Gets and sets output data after instrumentation process.
        /// </summary>
        public ObservableCollection<string> OutputData
        {
            get
            {
                if (_outputData == null)
                {
                    OutputData = new ObservableCollection<string>();
                }
                return _outputData;
            }
            set
            {
                _outputData = value;
                RaisePropertyChanged("OutputData");
            }
        }

        private Process _firstProcess;

        /// <summary>
        /// Gets and sets the first process of the instrumentation.
        /// </summary>
        public Process FirstProcess
        {
            get
            {
                if (_firstProcess == null)
                {
                    FirstProcess = new Process();
                }
                return _firstProcess;
            }
            set
            {
                _firstProcess = value;
                RaisePropertyChanged("FirstProcess");
            }
        }

        private Process _secondProcess;

        /// <summary>
        /// Gets and sets the second process of the instrumentation.
        /// </summary>
        public Process SecondProcess
        {
            get
            {
                if (_secondProcess == null)
                {
                    SecondProcess = new Process();
                }
                return _secondProcess;
            }
            set
            {
                _secondProcess = value;
                RaisePropertyChanged("SecondProcess");
            }
        }


        #region model reference

        TestCases _model;

        /// <summary>
        /// Gets and sets the model.
        /// </summary>
        public TestCases Model
        {
            get { return _model; }
            set { _model = value; }
        }

        #endregion
        #region states

        bool _firstScriptChosen;

        /// <summary>
        /// Gets and sets the flag for a chosen first script.
        /// </summary>
        public bool FirstScriptChosen
        {
            get { return _firstScriptChosen; }
            set
            {
                _firstScriptChosen = value;
                RaisePropertyChanged("FirstScriptChosen");
            }
        }

        bool _secondScriptChosen;

        /// <summary>
        /// Gets and sets the flag for a chosen second script.
        /// </summary>
        public bool SecondScriptChosen
        {
            get { return _secondScriptChosen; }
            set
            {
                _secondScriptChosen = value;
                RaisePropertyChanged("SecondScriptChosen");
            }
        }

        bool _apkChosen = false;

        /// <summary>
        /// Gets and sets the flag for a chosen apk.
        /// </summary>
        public bool ApkChosen
        {
            get { return _apkChosen; }
            set { _apkChosen = value; }
        }

        bool _ipEntered = false;

        /// <summary>
        /// Gets and sets the flag for enetred IP address.
        /// </summary>
        public bool IpEntered
        {
            get { return _ipEntered; }
            set { _ipEntered = value; }
        }

        //PACKAGE
        bool _packageChosen = false;

        /// <summary>
        /// Gets and sets the flag for a chosen package.
        /// </summary>
        public bool PackageChosen
        {
            get { return _packageChosen; }
            set { _packageChosen = value; }
        }

        #endregion

        #region contstructor

        /// <summary>
        /// Constructor for view model of code instrumentation.
        /// </summary>
        /// <param name="model"></param>
        public CodeInstrumentationViewModel(TestCases model)
        {
            _model = model;
        }

        private void runasync(object oparam)
        {
        }

        private void DoWork(object sender, DoWorkEventArgs e)
        {
            Application.Current.Dispatcher.Invoke(DispatcherPriority.Background, new NextPrimeDelegate(RunScript));
        }

        #endregion
        #region commands

        /// <summary>
        /// Method for ADT direectory choosing.
        /// </summary>
        /// <param name="parameter"></param>
        public void ChooseAdtDirectory(object parameter)
        {
            var result = OpenDirectoryDialog();
            if (!string.IsNullOrWhiteSpace(result))
            {
                Model.AdtDirectory = result;
            }
        }

        /// <summary>
        /// Method for checking if choosing ADT directory is possible.
        /// </summary>
        /// <param name="parameter"></param>
        /// <returns>
        /// Returns true if possible.
        /// </returns>
        public bool CanChooseAdtDirectory(object parameter)
        {
            return true;
        }

        /// <summary>
        /// Getting the suitable command.
        /// </summary>
        public ICommand ChooseAdtDirectoryCommand { get { return new RelayCommand<object>(ChooseAdtDirectory, CanChooseAdtDirectory); } }

        /// <summary>
        /// Checks whether choosing the first script is possible.
        /// </summary>
        /// <param name="parameter"></param>
        public void ChooseFirstScript(object parameter)
        {
            var result = OpenDialog("bat");
            if (!string.IsNullOrWhiteSpace(result))
            {
                Model.FirstScript = result;
            }
        }

        /// <summary>
        /// Defines whether choosing the first script is possible.
        /// </summary>
        /// <param name="parameter"></param>
        /// <returns>
        /// REturns true if possible.
        /// </returns>
        public bool CanChooseFirstScript(object parameter)
        {
            return true;
        }

        /// <summary>
        /// Getting the suitable command.
        /// </summary>
        public ICommand ChooseFirstScriptCommand { get { return new RelayCommand<object>(ChooseFirstScript, CanChooseFirstScript); } }

        /// <summary>
        /// Removes all error data from collection.
        /// </summary>
        /// <param name="parameter"></param>
        public void ClearErrors(object parameter)
        {
            ErrorData.Clear();
        }

        /// <summary>
        /// Defines whether clearing of all elements is possible.
        /// </summary>
        /// <param name="parameter"></param>
        /// <returns>
        /// Returns true if possible.
        /// </returns>
        public bool CanClearErrors(object parameter)
        {
            return true;
        }

        /// <summary>
        /// Getting the suitable command.
        /// </summary>
        public ICommand ClearErrorsCommand { get { return new RelayCommand<object>(ClearErrors, CanClearErrors); } }

        /// <summary>
        /// Removes all output data from collection.
        /// </summary>
        /// <param name="parameter"></param>
        public void ClearOutputs(object parameter)
        {
            OutputData.Clear();
        }

        /// <summary>
        /// Defines whether clearing of all elements is possible.
        /// </summary>
        /// <param name="parameter"></param>
        /// <returns>
        /// Returns true if possible. 
        /// </returns>
        public bool CanClearOutputs(object parameter)
        {
            return true;
        }

        /// <summary>
        /// Getting the suitable command.
        /// </summary>
        public ICommand ClearOutputsCommand { get { return new RelayCommand<object>(ClearOutputs, CanClearOutputs); } }

        /// <summary>
        /// Checks whether it is possible to choose apk file name.
        /// </summary>
        /// <param name="parameter"></param>
        public void ChooseApkFileName(object parameter)
        {
            var result = OpenDialog("apk");
            if (!string.IsNullOrWhiteSpace(result))
            {
                Model.ApkFileName = result;
            }
        }

        /// <summary>
        /// Defines whether choosing apk file name is possible.
        /// </summary>
        /// <param name="parameter"></param>
        /// <returns>
        /// Returns true if possible.
        /// </returns>
        public bool CanChooseApkFileName(object parameter)
        {
            return true;
        }

        /// <summary>
        /// Getting the suitable command.
        /// </summary>
        public ICommand ChooseApkFileNameCommand { get { return new RelayCommand<object>(ChooseApkFileName, CanChooseApkFileName); } }

        /// <summary>
        /// Checks whether choosing the second script is possible
        /// </summary>
        /// <param name="parameter"></param>
        public void ChooseSecondScript(object parameter)
        {
            var result = OpenDialog("bat");
            if (!string.IsNullOrWhiteSpace(result))
            {
                Model.SecondScript = result;
            }
        }

        /// <summary>
        /// Defines whether choosing the first script is possible.
        /// </summary>
        /// <param name="parameter"></param>
        /// <returns>
        /// Returns true if possible.
        /// </returns>
        public bool CanChooseSecondScript(object parameter)
        {
            return true;
        }

        /// <summary>
        /// Getting the suitable command.
        /// </summary>
        public ICommand ChooseSecondScriptCommand { get { return new RelayCommand<object>(ChooseSecondScript, CanChooseSecondScript); } }

        /// <summary>
        /// Declares delegate method. 
        /// </summary>
        public delegate void NextPrimeDelegate();

        /// <summary>
        /// Method for delegate execution asynchronously.
        /// </summary>
        /// <param name="param"></param>
        public void RunScriptAsync(object param)
        {
            Application.Current.Dispatcher.BeginInvoke(new Action(RunScript), DispatcherPriority.Background);
        }

        bool _coludRunScript = true;

        /// <summary>
        /// Gets and sets the possibility to run a script.
        /// </summary>
        public bool ColudRunScript
        {
            get { return _coludRunScript; }
            set
            {
                _coludRunScript = value;
                RaisePropertyChanged("ColudRunScript");
            }
        }

        /// <summary>
        /// Method that allows script run.
        /// </summary>
        public void RunScript()
        {
            ColudRunScript = false;
            Run = true;
            string args = string.Format("\"{0}\" {1} {2} {3}", Model.ApkFileName,
                                                                Path.GetFileNameWithoutExtension(Model.ApkFileName) + "_instrumented",
                                                                Model.AdtDirectory,
                                                                Model.AdtVersion);
            StartProcess(Model.FirstScript,
                            args,
                            firstScriptExited);
        }

        /// <summary>
        /// Informs if the script can be run.
        /// </summary>
        /// <param name="parameter"></param>
        /// <returns>
        /// Return notice of script running.
        /// </returns>
        public bool CanRunScript(object parameter)
        {
            return ColudRunScript;
        }

        /// <summary>
        /// Getting the suitable command.
        /// </summary>
        public ICommand RunScriptCommand { get { return new RelayCommand<object>(RunScriptAsync, CanRunScript); } }

        /// <summary>
        /// Method that tries to stop execution of scripts.
        /// </summary>
        /// <param name="pramaeter"></param>
        public void StopScript(object pramaeter)
        {
            try
            {
                FirstProcess.Kill();
                FirstProcess.Close();
            }
            catch { }
            try
            {
                SecondProcess.Kill();
                SecondProcess.Close();
            }
            catch { }
            Run = false;
            ColudRunScript = true;
        }

        /// <summary>
        /// Informs whether the script could stop.
        /// </summary>
        /// <param name="parameter"></param>
        /// <returns>
        /// Returns "impossible to run script" if the script could be stopped.
        /// </returns>
        public bool CanStopScript(object parameter)
        {
            return !ColudRunScript;
        }

        /// <summary>
        /// Getting the suitable command.
        /// </summary>
        public ICommand StopScriptCommand { get { return new RelayCommand<object>(StopScript, CanStopScript); } }

        #endregion

        #region methods

        #region delegates

        /// <summary>
        /// Declares delegate method. 
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        public delegate void ExitedDelegate(object sender, EventArgs e);

        /// <summary>
        /// Obtains file names from specified directory.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void firstScriptExited(object sender, EventArgs e)
        {            
            try
            {
                var csvFiles = Directory.GetFiles(Path.GetDirectoryName(Model.FirstScript) + "\\bin", "*.csv");
                foreach (var item in csvFiles)
                {
                    File.Copy(item, Model.MethodsDirectory + @"\" + Path.GetFileName(item));
                }
            }
            catch { }

            UpdatePermisinonInAndroidManifest(Path.GetDirectoryName(Model.FirstScript) + @"\CoverageProject\AndroidManifest.xml");

            var secArg = string.Format("{0} {1} {2} {3} {4} {5}", Model.HostAddress,
                                                                    Model.PackageName, Path.GetFileNameWithoutExtension(Model.ApkFileName) + "_instrumented.apk",
                                                                    Convert.ToInt32(Model.IsSystemApp),
                                                                    Model.AdtDirectory, Model.AdtVersion);
            SecondProcess = new Process();
            StartProcess(Model.SecondScript, secArg, secoundScriptExited);
        }

        /// <summary>
        /// Method for delegate execution asynchronously.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void secoundScriptExited(object sender, EventArgs e)
        {
            App.Current.Dispatcher.BeginInvoke(new Action(() => { EndSecondScript(); }));
        }

        private void EndSecondScript()
        {
            ColudRunScript = true;
            Run = false;
        }

        #endregion

        private void StartProcess(string path, string arguments, ExitedDelegate a)
        {
            var Runner = new Process();
            Runner.StartInfo.WorkingDirectory = Path.GetDirectoryName(path);
            Runner.StartInfo.FileName = path;
            Runner.StartInfo.Arguments = arguments;
            Runner.EnableRaisingEvents = true;

            Runner.StartInfo.CreateNoWindow = true;
            Runner.StartInfo.UseShellExecute = false;
            Runner.EnableRaisingEvents = true;

            Runner.StartInfo.RedirectStandardOutput = true;
            Runner.OutputDataReceived += Runner_OutputDataReceived;
            Runner.StartInfo.RedirectStandardInput = true;


            Runner.StartInfo.RedirectStandardError = true;
            Runner.ErrorDataReceived += Runner_ErrorDataReceived;
            Runner.Exited += new EventHandler(a);

            Runner.Start();
            Runner.BeginOutputReadLine();
            Runner.BeginErrorReadLine();

        }

        private bool _run;

        /// <summary>
        /// Gets and sets value for run.
        /// </summary>
        public bool Run
        {
            get { return _run; }
            set
            {
                _run = value;
                RaisePropertyChanged("Run");
            }
        }

        /// <summary>
        /// Method for delegate execution asynchronously.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void Runner_ErrorDataReceived(object sender, DataReceivedEventArgs e)
        {
            App.Current.Dispatcher.BeginInvoke(new Action(() => { ErrorData.Add(e.Data); }));
        }

        /// <summary>
        /// Method for delegate execution asynchronously.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void Runner_OutputDataReceived(object sender, DataReceivedEventArgs e)
        {
            App.Current.Dispatcher.BeginInvoke(new Action(() => { OutputData.Add(e.Data); }));
        }

        private void UpdatePermisinonInAndroidManifest(string filename)
        {
            UpdatePermisinonInAndroidManifest(filename, filename);
        }

        private static void UpdatePermisinonInAndroidManifest(string oldfilename, string newfilename)
        {
            if (File.Exists(oldfilename))
            {
                XmlDocument doc = new XmlDocument();
                doc.Load(oldfilename);
                XmlNode myNode = doc.DocumentElement.SelectSingleNode("uses-permission");
                doc.PreserveWhitespace = true;
                if (myNode != null)
                {
                    XmlAttribute attrib = myNode.Attributes["android:name"];

                    if (attrib != null)
                    {
                        if (attrib.Value == "android.permission.INTERNET")
                        {
                            return;
                        }
                        attrib.Value = "android.permission.INTERNET";
                    }
                    else
                    {
                        attrib = doc.CreateAttribute("name", "http://schemas.android.com/apk/res/android");
                        attrib.Value = "android.permission.INTERNET";
                    }
                }
                else
                {
                    XmlNode root = doc.DocumentElement;
                    XmlNode usesPermisionNode = doc.CreateElement("uses-permission");
                    var attrib = doc.CreateAttribute("name", "http://schemas.android.com/apk/res/android");
                    attrib.Value = "android.permission.INTERNET";
                    usesPermisionNode.Attributes.Append(attrib);
                    root.InsertAfter(usesPermisionNode, root.FirstChild);
                }
                doc.Save(newfilename);
                using (XmlTextWriter writer = new XmlTextWriter(newfilename, null))
                {
                    writer.Formatting = Formatting.Indented;
                    doc.Save(writer);
                }
            }
            else
            {
                System.Windows.Forms.MessageBox.Show("AndroidManifest.xml does not exist!");
            }
        }

        #endregion

        #region Dialogs

        private string OpenDialog(string extension)
        {
            var filename = "";
            var dialog = new OpenFileDialog();
            dialog.Filter = string.Format("{0} (*.{0})|*.{0}", extension);
            dialog.ValidateNames = true;
            dialog.CheckPathExists = true;
            var showed = dialog.ShowDialog();
            if (showed != null && showed == true)
            {
                filename = dialog.FileName;
            }
            return filename;
        }

        private string OpenDirectoryDialog()
        {
            var folderdialog = new System.Windows.Forms.FolderBrowserDialog();
            folderdialog.RootFolder = System.Environment.SpecialFolder.ProgramFiles; //@"c:\Program files\Rt-Executor"
            folderdialog.ShowDialog();
            return folderdialog.SelectedPath.ToString();
        }

        #endregion
    }
}
